# Description: Distance to closest record metric
# Author: Anton D. Lautrup
# Date: 23-08-2023

import numpy as np

from ..core.metric import MetricClass

from ...utils.nn_distance import _knn_distance

class MedianDistanceToClosestRecord(MetricClass):
    """The Metric Class is an abstract class that interfaces with 
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    self.analysis_target: variable name
    """

    def name() -> str:
        """ Name/keyword to reference the metric"""
        return 'dcr'

    def type() -> str:
        """ Set to 'privacy' or 'utility' """
        return 'privacy'

    def evaluate(self) -> float | dict:
        """Distance to closest record, using the same NN stuff as NNAA"""
        
        distances = _knn_distance(self.synt_data,self.real_data,self.cat_cols,1,self.nn_dist)
        in_dists = _knn_distance(self.real_data,self.real_data,self.cat_cols,1,self.nn_dist)

        int_nn = np.median(in_dists)
        mut_nn = np.median(distances)

        if (int_nn == 0 and mut_nn == 0): dcr = 1
        elif (int_nn == 0 and mut_nn != 0): dcr = 0
        else: dcr = mut_nn/int_nn
        self.results = {'mDCR': dcr}
        return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval. 
|                                          :                    |"""
        string = """\
| Median distance to closest record        :   %.4f           |""" % (self.results['mDCR'])
        return string

    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).
        
        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0

        Error fields can be empty.
        """
        if self.results != {}:
            return [{'metric': 'median_DCR', 'dim': 'p', 
                     'val': self.results['mDCR'],
                     'n_val': np.tanh(self.results['mDCR']),
                    #  'idx_val': np.tanh(self.results['mDCR'])
                     }]
        else: pass


# Description: Epsilon identifiability metric
# Author: Anton D. Lautrup
# Date: 23-08-2023

import numpy as np

from ..core.metric import MetricClass

from scipy.stats import entropy

from ...utils.nn_distance import _knn_distance

def _column_entropy(labels):
        value, counts = np.unique(np.round(labels), return_counts=True)
        return entropy(counts)

class EpsilonIdentifiability(MetricClass):
    """The Metric Class is an abstract class that interfaces with 
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    self.analysis_target: variable name

    """

    def name() -> str:
        """ Name/keyword to reference the metric"""
        return 'eps_risk'

    def type() -> str:
        """ Set to 'privacy' or 'utility' """
        return 'privacy'

    def evaluate(self) -> float | dict:
        """Function for computing the epsilon identifiability risk

        Adapted from:
        Yoon, J., Drumright, L. N., & van der Schaar, M. (2020). Anonymization Through Data Synthesis Using Generative Adversarial Networks (ADS-GAN). 
        IEEE Journal of Biomedical and Health Informatics, 24(8), 2378–2388. [doi:10.1109/JBHI.2020.2980262] 
        """

        if self.nn_dist == 'euclid':
            self.real_data = self.real_data[self.num_cols]
            self.synt_data = self.synt_data[self.num_cols] 

        real = np.asarray(self.real_data)

        no, x_dim = np.shape(real)
        W = [_column_entropy(real[:, i]) for i in range(x_dim)]
        W_adjust = 1/(np.array(W)+1e-16)

        # for i in range(x_dim):
        #     real_hat[:, i] = real[:, i] * 1. / W[i]
        #     fake_hat[:, i] = fake[:, i] * 1. / W[i]

        in_dists = _knn_distance(self.real_data,self.real_data,self.cat_cols,1,self.nn_dist,W_adjust)[0]
        ext_distances = _knn_distance(self.real_data,self.synt_data,self.cat_cols,1,self.nn_dist,W_adjust)[0]

        R_Diff = ext_distances - in_dists
        identifiability_value = np.sum(R_Diff < 0) / float(no)

        self.results['eps_risk'] = identifiability_value

        if self.hout_data is not None:
            in_dists = _knn_distance(self.hout_data,self.hout_data,self.cat_cols,1,self.nn_dist,W_adjust)[0]
            ext_distances = _knn_distance(self.hout_data,self.synt_data,self.cat_cols,1,self.nn_dist,W_adjust)[0]

            R_Diff = ext_distances - in_dists
            identifiability_value = np.sum(R_Diff < 0) / float(no)

            self.results['priv_loss'] = self.results['eps_risk'] - identifiability_value

        return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval. 
|                                          :                    |"""
        string = """\
| Epsilon identifiability risk             :   %.4f           |""" % (self.results['eps_risk'])
        if (self.results != {} and self.hout_data is not None):
             string += """       
| Privacy loss (diff. in eps. risk)        :   %.4f           |""" % (self.results['priv_loss'])
        return string

    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).

        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0
        """
        if self.results != {}:
            output =  [{'metric': 'eps_identif_risk', 'dim': 'p', 
                     'val': self.results['eps_risk'], 
                     'n_val': 1-self.results['eps_risk'], 
                     }]
            if self.hout_data is not None:
                output.extend([{'metric': 'priv_loss_eps', 'dim': 'p', 
                        'val': self.results['priv_loss'], 
                        'n_val': 1-abs(self.results['priv_loss']), 
                        }])
            return output
        else: pass



# Description: Membership Inference Attack based on classification
# Author: Tobias Hyrup
# Date: 2023-10-30

import numpy as np
import pandas as pd
from logging import warning
from ..core.metric import MetricClass
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import precision_score, recall_score, f1_score
from sklearn.model_selection import train_test_split


class MIAClassifier(MetricClass):
    """The Metric Class is an abstract class that interfaces with
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    self.analysis_target: variable name

    """

    def name() -> str:
        """Name/keyword to reference the metric"""
        return "mia_risk"

    def type() -> str:
        """Set to 'privacy' or 'utility'"""
        return "privacy"

    def evaluate(self, num_eval_iter=5) -> float | dict:
        """Function for computing the precision, recall, and F1-score of a membership inference attack using a Random Forest classifier"""
        try:
            assert self.hout_data is not None
        except AssertionError:
            print(" Warning: Membership inference attack metric did not run, holdout data was not supplied!")
            pass
        else:
            if len(self.real_data) < len(self.hout_data) // 2:
                    warning(
                        "The holdout data is more than double the size of the real data. The holdout data will be downsampled to match the size of the real data. real size: %s, holdout size: %s", len(self.real_data), len(self.hout_data)
                    )
            # One-hot encode. All data is combined to ensure consitent encoding
            combined_data = pd.concat(
                [self.real_data, self.synt_data, self.hout_data], ignore_index=True
            )
            combined_data_encoded = pd.get_dummies(
                combined_data, columns=self.cat_cols, drop_first=True
            )

            # Separate into the three datasets
            real = combined_data_encoded.iloc[: len(self.real_data)]
            syn = combined_data_encoded.iloc[
                len(self.real_data) : len(self.real_data) + len(self.synt_data)
            ]
            hout = combined_data_encoded.iloc[
                len(self.real_data) + len(self.synt_data) :
            ]

            # Run classifier multiple times and average the results
            pre_results = {
                "precision": [],
                "recall": [],
                "f1": [],
            }
            for _ in range(num_eval_iter):
                hout_train, hout_test = train_test_split(hout, test_size=0.5)

                # Create training data consisting of synthetic and holdout data
                X_train = pd.concat([syn, hout_train], axis=0)
                y_train = pd.Series([1] * len(syn) + [0] * len(hout_train))
                
                # Create test set by combining some random data from the real and holdout data with an equal number of records from each dataframe
                if len(real) < len(hout_test):
                    # warning(
                    #     "The holdout data is larger than the real data. The holdout data will be downsampled to match the size of the real data."
                    # )
                    hout_sample = hout_test.sample(n=len(real))
                    real_sample = real
                else:
                    real_sample = real.sample(n=len(hout_test))
                    hout_sample = hout_test
                X_test = pd.concat(
                    [
                        real_sample,
                        hout_sample,
                    ],
                    axis=0,
                )
                y_test = pd.Series([1] * len(real_sample) + [0] * len(hout_sample))

                # Train the classifier on all the data
                rf_classifier = RandomForestClassifier(n_estimators=100).fit(
                    X_train, y_train
                )

                # Get predictions
                holdout_predictions = rf_classifier.predict(X_test)

                # Calculate precision, recall, and F1-score
                pre_results["precision"].append(
                    precision_score(y_test, holdout_predictions)
                )
                pre_results["recall"].append(recall_score(y_test, holdout_predictions))
                pre_results["f1"].append(
                    f1_score(y_test, holdout_predictions, average="macro")
                )

            precision = np.mean(pre_results["precision"])
            precision_se = np.std(pre_results["precision"], ddof=1) / np.sqrt(
                num_eval_iter
            )

            recall = np.mean(pre_results["recall"])
            recall_se = np.std(pre_results["recall"], ddof=1) / np.sqrt(num_eval_iter)

            f1 = np.mean(pre_results["f1"])
            f1_se = np.std(pre_results["f1"], ddof=1) / np.sqrt(num_eval_iter)

            self.results = {
                "MIA precision": precision,
                "MIA precision se": precision_se,
                "MIA recall": recall,
                "MIA recall se": recall_se,
                "MIA macro F1": f1,
                "MIA macro F1 se": f1_se,
            }

            return self.results

    def format_output(self) -> str:
        """Return string for formatting the output, when the
                metric is part of SynthEval.
        |                                          :                    |"""
        try:
            assert self.hout_data is not None
        except AssertionError:
            pass
        else:
            string = """\
| Membership inference attack Classifier F1:   %.4f  %.4f   |
|   -> Precision                           :   %.4f  %.4f   |
|   -> Recall                              :   %.4f  %.4f   |""" % (
                self.results["MIA macro F1"],
                self.results["MIA macro F1 se"],
                self.results["MIA precision"],
                self.results["MIA precision se"],
                self.results["MIA recall"],
                self.results["MIA recall se"],
            )
            return string

    def normalize_output(self) -> list:
        """This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).

        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0
        """
        if self.results != {}:
            return [{
                    "metric": "mia_cls_risk",
                    "dim": "p",
                    "val": self.results["MIA macro F1"],
                    "err": self.results["MIA macro F1 se"],
                    "n_val": 1 - self.results["MIA macro F1"],
                    "n_err": self.results["MIA macro F1 se"],
                }]
        else: pass



# Description: Template script for making new metric classes
# Author: Anton D. Lautrup
# Date: 21-08-2023

from ..core.metric import MetricClass
import numpy as np

class MetricClassName(MetricClass):
    """The Metric Class is an abstract class that interfaces with 
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    self.analysis_target: variable name

    self.verbose   : bool (mainly for supressing prints and plots)

    """

    def name() -> str:
        """ Name/keyword to reference the metric"""
        return 'q_mse'

    def type() -> str:
        """ Set to 'privacy' or 'utility' """
        return 'utility'

    def evaluate(self, num_quants=10, cat_mse=False) -> float | dict:
        """Function for executing the quantile mse metric.

        Args:
            num_quants: int - number of quantiles to evaluate
            cat_mse: bool - whether to calculate the mse for categorical variables
                                (has nothing to do with quantiles, hence the option)
    
        Returns:
            Avg. qMSE: dict  - holds avg. and standard error of the mean (SE)
        """
        try:
            assert (len(self.num_cols)>=1 or cat_mse)
        except AssertionError:
            print('Error: Quantile mse did not run, no nummerical attributes, or cat_mse not enabled!')
        else:
            qMSE_lst = []
            for category in self.real_data.columns:

                if category in self.cat_cols and cat_mse:
                    # Categorical data
                    real_items = self.real_data[category].unique()

                    synth_frac = np.array([np.sum(self.synt_data[category] == item) for item in real_items]) / len(self.synt_data)
                    real_frac = np.array([np.sum(self.real_data[category] == item) for item in real_items]) / len(self.real_data)

                    qMSE_lst.append(np.mean((synth_frac - real_frac)**2))
                else:
                    # Numerical data
                    quantiles = np.quantile(self.real_data[category], np.linspace(0, 1, num_quants+1))
                    bin_edges = quantiles.tolist()

                    synth_hist, _ = np.histogram(self.synt_data[category], bins=bin_edges)
                    synth_frac = synth_hist / len(self.synt_data)

                    qMSE_lst.append(np.mean((synth_frac - 1/num_quants)**2))
            
            self.results = {'avg qMSE': np.mean(qMSE_lst), 
                            'qMSE err': np.std(qMSE_lst, ddof=1) / np.sqrt(len(qMSE_lst))
                            }
            return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval. 
|                                          :                    |"""
        if self.results != {}:
            string = """\
| Quantile mean squared error (qMSE)       :   %.4f  %.4f   |""" % (
            self.results['avg qMSE'],
            self.results['qMSE err'],
            )
            return string
        else: pass

    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).

        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0
        """
        if self.results != {}:
            return [{'metric': 'avg_qMSE', 'dim': 'u', 
                     'val': self.results['avg qMSE'], 
                     'err': self.results['qMSE err'], 
                     'n_val': 1-self.results['avg qMSE'], 
                     'n_err': self.results['qMSE err'], 
                     }]
        else: pass



# Description: Confidence interval overlap
# Author: Anton D. Lautrup
# Date: 23-08-2023

import numpy as np

from scipy.stats import sem

from ..core.metric import MetricClass

class ConfidenceIntervalOverlap(MetricClass):
    """The Metric Class is an abstract class that interfaces with 
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    self.analysis_target: variable name
    """

    def name() -> str:
        """ Name/keyword to reference the metric"""
        return 'cio'

    def type() -> str:
        """ Set to 'privacy' or 'utility' """
        return 'utility'

    def evaluate(self,confidence=95) -> float | dict:
        """Function for calculating the average CIO, also returns the 
        number of nonoverlapping interval"""
        confidence_table = {80: 1.28, 90: 1.645, 95: 1.96, 98: 2.33, 99: 2.58}
        try:
            assert len(self.num_cols) > 0
            assert confidence in confidence_table.keys()
        except AssertionError:
            if len(self.num_cols) == 0:
                print(" Warning: No nummerical attributes provided for confidence interval overlap metric.")
            else:
                print(" Error: Confidence level not recognized, choose 80, 90, 95, 98 or 99.")
            return {}
        else:
            self.confidence = confidence

            if confidence in confidence_table.keys():
                z_value = confidence_table[confidence]
                mus = np.array([np.mean(self.real_data[self.num_cols],axis=0),np.mean(self.synt_data[self.num_cols],axis=0)]).T
                sems = np.array([sem(self.real_data[self.num_cols]),sem(self.synt_data[self.num_cols])]).T
                
                CI = sems*z_value
                us = mus+sems
                ls = mus-sems

                Jk = []
                for i in range(len(CI)):
                    top = (min(us[i][0],us[i][1])-max(ls[i][0],ls[i][1]))
                    Jk.append(max(0,0.5*(top/(us[i][0]-ls[i][0])+top/(us[i][1]-ls[i][1]))))

                num = sum([j == 0 for j in Jk])
                frac = num/len(Jk)
                self.results = {'avg overlap': np.mean(Jk), 
                                'overlap err': np.std(Jk,ddof=1)/np.sqrt(len(Jk)), 
                                'num non-overlaps': num, 
                                'frac non-overlaps': frac
                                }
                return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval.
|                                          :                    |"""
        if self.results != {}:
            string = """\
| Average confidence interval overlap      :   %.4f  %.4f   |
|   -> # non-overlapping COIs at %2d%%       :   %2d               |
|   -> fraction of non-overlapping CIs     :   %.4f           |""" % (
            self.results['avg overlap'],
            self.results['overlap err'],
            self.confidence,
            self.results['num non-overlaps'],
            self.results['frac non-overlaps'])
            return string
        else: pass

    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).

        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0
        """
        if self.results != {}:
            # power = np.exp(-8*(self.results['avg overlap']-0.5))
            # val_non_lin     = 1/(1+power)
            # val_non_lin_err = 8*power/((1+power)**2)*self.results['overlap err']
            return [{'metric': 'avg_cio', 'dim': 'u', 
                     'val': self.results['avg overlap'], 
                     'err': self.results['overlap err'], 
                     'n_val': self.results['avg overlap'], 
                     'n_err': self.results['overlap err'], 
                    #  'idx_val': val_non_lin, 
                    #  'idx_err': val_non_lin_err
                     }]
        else: pass


# Description: Nearest neighbour adversarial accuracy implementation
# Author: Anton D. Lautrup
# Date: 21-08-2023

import numpy as np

from ..core.metric import MetricClass

from ...utils.nn_distance import _knn_distance

def _adversarial_score(real, fake, cat_cols, metric):
    """Function for calculating adversarial score"""
    left = np.mean(_knn_distance(real, fake, cat_cols, 1, metric)[0] > _knn_distance(real, real, cat_cols, 1, metric)[0])
    right = np.mean(_knn_distance(fake, real, cat_cols, 1, metric)[0] > _knn_distance(fake, fake, cat_cols, 1, metric)[0])
    return 0.5 * (left + right)

def evaluate_dataset_nnaa(real, fake, num_cols, cat_cols, metric, n_resample):
    """Helper function for running adversarial score multiple times if the 
    datasets have much different sizes."""

    real_fake = len(real)/len(fake)
    fake_real = len(fake)/len(real)

    if any([real_fake >= 2, fake_real >= 2]):
        aa_lst = []
        for batch in range(n_resample):
            temp_r = real if real_fake < 2 else real.sample(n=len(fake))
            temp_f = fake if fake_real < 2 else fake.sample(n=len(real))
            aa_lst.append(_adversarial_score(temp_r, temp_f, cat_cols, metric))

        avg = np.mean(aa_lst)
        err = np.std(aa_lst, ddof=1)/np.sqrt(len(aa_lst))
    else:
        avg = _adversarial_score(real, fake, cat_cols, metric)
        err = 0.0

    return avg, err

class NearestNeighbourAdversarialAccuracy(MetricClass):
    """The Metric Class is an abstract class that interfaces with 
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings
    
    self.nn_dist   : string keyword
    
    """

    def name() -> str:
        """ Name/keyword to reference the metric"""
        return 'nnaa'

    def type() -> str:
        """ Set to 'privacy' or 'utility' """
        return 'utility'

    def evaluate(self, n_resample=30) -> dict:
        """Implementation heavily inspired by original paper"""

        avg, err = evaluate_dataset_nnaa(self.real_data,self.synt_data,self.num_cols,self.cat_cols,self.nn_dist,n_resample)

        self.results = {'avg': avg, 'err': err}

        if self.hout_data is not None:
            avg_h, err_h = evaluate_dataset_nnaa(self.hout_data,self.synt_data,self.num_cols,self.cat_cols,self.nn_dist,n_resample)
            diff = avg_h - avg
            err_diff = np.sqrt(err_h**2+err**2)

            self.results['priv_loss'] = diff
            self.results['priv_loss_err'] = err_diff

        return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval. 
|                                          :                    |"""
        string = """\
| Nearest neighbour adversarial accuracy   :   %.4f  %.4f   |""" % (self.results['avg'], self.results['err'])
        return string

    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).

        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0 
        """
        if self.results != {}:
            output =  [{'metric': 'nnaa', 'dim': 'u', 
                        'val': self.results['avg'], 
                        'err': self.results['err'], 
                        'n_val': 1-self.results['avg'], 
                        'n_err': self.results['err'], 
                        }]
            if self.hout_data is not None:
                output.extend([{'metric': 'priv_loss_nnaa', 'dim': 'p', 
                        'val': self.results['priv_loss'], 
                        'err': self.results['priv_loss_err'], 
                        'n_val': 1-abs(self.results['priv_loss']), 
                        'n_err': self.results['priv_loss_err'], 
                        }])
            return output
        else: pass

    def extra_formatted_output(self) -> str:
        """Bit for printing the privacy loss together with the other privacy metrics"""
        if (self.results != {} and self.hout_data is not None):
            string = """\
| Privacy loss (diff. in NNAA)             :   %.4f  %.4f   |""" % (self.results['priv_loss'], self.results['priv_loss_err'])
            return string
        else: pass



# Description: Hitting rate metric
# Author: Anton D. Lautrup
# Date: 23-08-2023

import numpy as np
from ..core.metric import MetricClass

class HittingRate(MetricClass):
    """The Metric Class is an abstract class that interfaces with 
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    self.analysis_target: variable name

    """

    def name() -> str:
        """ Name/keyword to reference the metric"""
        return 'hit_rate'

    def type() -> str:
        """ Set to 'privacy' or 'utility' """
        return 'privacy'

    def evaluate(self,thres_percent=1/30) -> float | dict:
        """For hitting rate we regard records as similar if the 
        nummerical attributes are within a threshold range(att)/30"""
        self.thres_percent = thres_percent
        thres = thres_percent*(self.real_data.max() - self.real_data.min())
        thres[self.cat_cols] = 0

        hit = 0
        for i, r in self.real_data.iterrows():
            hit += any((abs(r-self.synt_data) <= thres).all(axis='columns'))
        hit_rate = hit/len(self.real_data)
        self.results = {'hit rate': hit_rate}
        return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval. 
|                                          :                    |"""
        string = """\
| Hitting rate (%.2f x range(att))         :   %.4f           |""" % (self.thres_percent, self.results['hit rate'])
        return string
    
    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).
     
        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0
        """
        if self.results != {}:
            # val_non_lin = np.exp(-5*self.results['hit rate'])
            return [{'metric': 'hit_rate', 'dim': 'p', 
                     'val': self.results['hit rate'], 
                     'n_val': 1-self.results['hit rate'], 
                    #  'idx_val': val_non_lin, 
                     }]
        else: pass



# Description: Script for calculating the mixed correlation
# Author: Anton D. Lautrup
# Date: 23-08-2023

import numpy as np
import pandas as pd

from ..core.metric import MetricClass

from scipy.stats import chi2_contingency
from ...utils.plot_metrics import plot_matrix_heatmap

def _cramers_V(var1,var2) :
    """function for calculating Cramers V between two categorial variables
    credit: https://www.kaggle.com/code/chrisbss1/cramer-s-v-correlation-matrix
    """
    crosstab =np.array(pd.crosstab(var1, var2, rownames=None, colnames=None)) # Cross table building
    stat = chi2_contingency(crosstab)[0] # Keeping of the test statistic of the Chi2 test
    obs = np.sum(crosstab) # Number of observations
    mini = min(crosstab.shape)-1 # Take the minimum value between the columns and the rows of the cross table
    return (stat/(obs*mini+1e-16))

def _apply_mat(data,func,labs1,labs2):
    """Help function for constructing a matrix based on func accross labels 1 and 2"""
    res = (func(data[lab1],data[lab2]) for lab1 in labs1 for lab2 in labs2)
    return pd.DataFrame(np.fromiter(res, dtype=float).reshape(len(labs1),len(labs2)), columns = labs2, index = labs1)

def _correlation_ratio(categories, measurements):
    """Function for calculating the correlation ration eta^2 of categorial and nummerical data"""
    fcat, _ = pd.factorize(categories)
    cat_num = np.max(fcat)+1
    y_avg_array = np.zeros(cat_num)
    n_array = np.zeros(cat_num)
    for i in range(0,cat_num):
        cat_measures = measurements[np.argwhere(fcat == i).flatten()]
        n_array[i] = len(cat_measures)
        y_avg_array[i] = np.average(cat_measures)
    y_total_avg = np.sum(np.multiply(y_avg_array,n_array))/np.sum(n_array)
    numerator = np.sum(np.multiply(n_array,np.power(np.subtract(y_avg_array,y_total_avg),2)))
    denominator = np.sum(np.power(np.subtract(measurements,y_total_avg),2))
    if numerator == 0:
        eta = 0.0
    else:
        eta = numerator/denominator
    return eta

def mixed_correlation(data,num_cols,cat_cols):
    """Function for calculating a correlation matrix of mixed datatypes.
    Spearman's rho is used for rank-based correlation, Cramer's V is used for categorical variables, 
    and correlation ratio is used for categorical and continuous variables.
    """
    corr_num_num = data[num_cols].corr()
    corr_cat_cat = _apply_mat(data,_cramers_V,cat_cols,cat_cols)
    corr_cat_num = _apply_mat(data,_correlation_ratio,cat_cols,num_cols)
    if corr_cat_cat.empty: corr = corr_num_num
    elif corr_num_num.empty: corr = corr_cat_cat
    else:
        top_row = pd.concat([corr_cat_cat,corr_cat_num],axis=1)
        bot_row = pd.concat([corr_cat_num.transpose(),corr_num_num],axis=1)
        corr = pd.concat([top_row,bot_row],axis=0)
    return corr + np.diag(1-np.diag(corr))

class MixedCorrelation(MetricClass):
    """The Metric Class is an abstract class that interfaces with 
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    self.analysis_target: variable name
    """

    def name() -> str:
        """ Name/keyword to reference the metric"""
        return 'corr_diff'

    def type() -> str:
        """ Set to 'privacy' or 'utility' """
        return 'utility'

    def evaluate(self, mixed_corr=True, return_mats=False, axs_lim=(-1,1), axs_scale="RdBu") -> dict:
        """Function for calculating the (mixed) correlation matrix difference.
        This calculation uses spearmans rho for numerical-numerical, Cramer's V for categories,
        and correlation ratio (eta) for numerical-categorials.
        
        Mixed mode can be disabled, to only use the numerical variables."""
        self.mixed_corr = mixed_corr
        if mixed_corr:
            r_corr = mixed_correlation(self.real_data,self.num_cols,self.cat_cols)
            f_corr = mixed_correlation(self.synt_data,self.num_cols,self.cat_cols)
            corr_mat = r_corr-f_corr
            if self.verbose: plot_matrix_heatmap(corr_mat,'Mixed correlation matrix difference', 'corr', axs_lim, axs_scale)
        else:
            r_corr = self.real_data[self.num_cols].corr()
            f_corr = self.synt_data[self.num_cols].corr()
            corr_mat = r_corr-f_corr
            if self.verbose: plot_matrix_heatmap(corr_mat,'Correlation matrix difference (nums only)', 'corr', axs_lim, axs_scale)
        
        self.results = {'corr_mat_diff': np.linalg.norm(corr_mat,ord='fro'), 'corr_mat_dims': len(corr_mat)}
        if return_mats: self.results['real_cor_mat'] = r_corr
        if return_mats: self.results['synt_cor_mat'] = f_corr
        if return_mats: self.results['diff_cor_mat'] = corr_mat
        return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval. 
|                                          :                    |"""
        if self.mixed_corr:
            string = """\
| Mixed correlation matrix difference      :   %.4f           |""" % (self.results['corr_mat_diff'])
        else:
            string = """\
| Correlation difference (nums only)       :   %.4f           |""" % (self.results['corr_mat_diff'])
        return string

    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).

        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0  
        """
        if self.results != {}:
            n_elements = int(self.results['corr_mat_dims']*(self.results['corr_mat_dims']-1)/2)
            return [{'metric': 'corr_mat_diff', 'dim': 'u', 
                     'val': self.results['corr_mat_diff'], 
                     'n_val': 1-self.results['corr_mat_diff']/n_elements, 
                    #  'idx_val': 1-np.tanh(self.results['corr_mat_diff'])
                     }]
        else: pass



# Description: Hellinger distance metric class
# Author: Anton D. Lautrup
# Date: 18-08-2023

import numpy as np

from ..core.metric import MetricClass

def _scott_ref_rule(set1,set2):
    """Function for doing the Scott reference rule to calcualte number of bins needed to 
    represent the nummerical values."""
    samples = np.concatenate((set1, set2))
    std = np.std(samples)
    n = len(samples)
    bin_width = np.ceil(n**(1/3) * std / (3.5 * (np.percentile(samples, 75) - np.percentile(samples, 25)))).astype(int)

    min_edge = min(samples); max_edge = max(samples)
    N = min(abs(int((max_edge-min_edge)/bin_width)),10000); Nplus1 = N + 1
    return np.linspace(min_edge, max_edge, Nplus1)

def _hellinger(p,q):
    """Hellinger distance between distributions"""
    sqrt_pdf1 = np.sqrt(p)
    sqrt_pdf2 = np.sqrt(q)
    diff = sqrt_pdf1 - sqrt_pdf2
    return 1/np.sqrt(2)*np.linalg.norm(diff)

class HellingerDistance(MetricClass):

    def name() -> str:
        """name/keyword to reference the metric"""
        return 'h_dist'

    def type() -> str:
        """privacy or utility"""
        return 'utility'

    def evaluate(self) -> float | dict:
        """ Function for evaluating the metric"""
        H_dist = []
    
        for category in self.cat_cols:
            class_num = len(np.unique(self.real_data[category]))

            pdfR = np.histogram(self.real_data[category], bins=class_num)[0]
            pdfF = np.histogram(self.synt_data[category], bins=class_num)[0]
            H_dist.append(_hellinger(pdfR/sum(pdfR),pdfF/sum(pdfF)))
        
        for category in self.num_cols:
            n_bins = _scott_ref_rule(self.real_data[category],self.synt_data[category]) # Scott rule for finding bin width

            pdfR = np.histogram(self.real_data[category], bins=n_bins)[0]
            pdfF = np.histogram(self.synt_data[category], bins=n_bins)[0]
            H_dist.append(_hellinger(pdfR/sum(pdfR),pdfF/sum(pdfF)))

        self.results = {'avg': np.mean(H_dist), 'err': np.std(H_dist,ddof=1)/np.sqrt(len(H_dist))}
        return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval.        
        """
        string = """\
| Average empirical Hellinger distance     :   %.4f  %.4f   |""" % (self.results['avg'], self.results['err'])
        return string

    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).

        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0
        """
        if self.results != {}:
            # power = np.exp(10*(self.results['avg']-0.25))
            # val_non_lin     = 1/(1+power)
            # val_non_lin_err = 10*power/((1+power)**2)*self.results['err']

            return [{'metric': 'avg_h_dist', 'dim': 'u', 
                     'val': self.results['avg'], 
                     'err': self.results['err'], 
                     'n_val': 1-self.results['avg'], 
                     'n_err': self.results['err'], 
                    #  'idx_val': val_non_lin, 
                    #  'idx_err': val_non_lin_err
                     }]
        else: pass



# Description: Script for post-processing results for the benchmark view
# Author: Anton D. Lautrup
# Date: 13-12-2023

import numpy as np
import pandas as pd

from sklearn.preprocessing import MinMaxScaler

def summation_ranking(dataframe, utility_mets, privacy_mets):
    """Take the basic sum of the utility and privacy metrics as the rank."""
    res_df = dataframe.copy()

    res_df['rank'] = res_df.sum(axis=1)
    res_df['u_rank'] = res_df[utility_mets].sum(axis=1)
    res_df['p_rank'] = res_df[privacy_mets].sum(axis=1)
    return res_df

def linear_ranking(dataframe, utility_mets, privacy_mets):
    """Apply minmax scaling to the normalised columns, take the row sum as rank.
    Appropriate when there is enough separation between scores that we can trust 
    a linear scale like this."""
    scaler = MinMaxScaler()
    res = scaler.fit_transform(dataframe)
    res_df = pd.DataFrame(res,columns=dataframe.columns,index=dataframe.index)

    res_df['rank'] = res_df.sum(axis=1,numeric_only = True)
    res_df['u_rank'] = res_df[utility_mets].sum(axis=1,numeric_only = True)
    res_df['p_rank'] = res_df[privacy_mets].sum(axis=1,numeric_only = True)
    return res_df

def extremes_ranking(dataframe, utility_mets, privacy_mets):
    """Map worst and best score to 0 and 1 respectively, everything else is 0.5.
    This scheme works to separate overall best and worst from normally distributed 
    mass, where we may not be able to say much objectively founded about the
    intermediate results subject to noise."""
    res_df = dataframe.copy()
    res_df[:] = 0.5

    minimum = dataframe.astype(float).min(axis=0)
    maximum = dataframe.astype(float).max(axis=0)
    
    for col in res_df.columns:
        min_con = dataframe[col] == minimum[col]
        max_con = dataframe[col] == maximum[col]

        res_df[col].mask(min_con,0,inplace=True)
        res_df[col].mask(max_con,1,inplace=True)

    res_df['rank'] = res_df.sum(axis=1)
    res_df['u_rank'] = res_df[utility_mets].sum(axis=1)
    res_df['p_rank'] = res_df[privacy_mets].sum(axis=1)
    return res_df

def quantile_ranking(dataframe, utility_mets, privacy_mets):
    """Expermental: Use only if you have enough samples! Sort the results into
    four quantiles and score them 0, 1, 2, 3. Appropriate for lots of samples that
    are not all on top of each other, e.g. high variance, uniform distribution etc."""

    res_df = dataframe.astype(float).apply(lambda x: pd.qcut(x, 4, labels=False, duplicates='drop'))
    
    res_df['rank'] = res_df.sum(axis=1)
    res_df['u_rank'] = res_df[utility_mets].sum(axis=1)
    res_df['p_rank'] = res_df[privacy_mets].sum(axis=1)
    return res_df



# Description: Nearest neighbour distance ratio metric
# Author: Anton D. Lautrup
# Date: 23-08-2023

import numpy as np

from ..core.metric import MetricClass

from ...utils.nn_distance import _knn_distance

class NearestNeighbourDistanceRatio(MetricClass):
    """The Metric Class is an abstract class that interfaces with 
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    self.analysis_target: variable name

    """

    def name() -> str:
        """ Name/keyword to reference the metric"""
        return 'nndr'

    def type() -> str:
        """ Set to 'privacy' or 'utility' """
        return 'privacy'

    def evaluate(self) -> dict:
        """ Compute the Nearest Neighbour Distance Ratio (NNDR) between two datasets.
        """
        dist = _knn_distance(self.real_data, self.synt_data, self.cat_cols, 2, self.nn_dist)
        dr = list(map(lambda x: x[0] / (x[1]+1e-16), np.transpose(dist)))

        self.results = {'avg': np.mean(dr), 'err': np.std(dr,ddof=1)/np.sqrt(len(dr))}

        if self.hout_data is not None:
            dist_h = _knn_distance(self.hout_data, self.synt_data, self.cat_cols, 2, self.nn_dist)
            dr_h = list(map(lambda x: x[0] / (x[1]+1e-16), np.transpose(dist_h)))
            diff     = np.mean(dr_h) - self.results['avg']
            err_diff = np.sqrt((np.std(dr_h,ddof=1)/np.sqrt(len(dr_h)))**2+self.results['err']**2)

            self.results['priv_loss'] = diff
            self.results['priv_loss_err'] = err_diff
        return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval. 
|                                          :                    |"""
        string = """\
| Nearest neighbour distance ratio         :   %.4f  %.4f   |""" % (self.results['avg'], self.results['err'])
        if (self.results != {} and self.hout_data is not None):
             string += """\n\
| Privacy loss (diff. in NNDR)             :   %.4f  %.4f   |""" % (self.results['priv_loss'], self.results['priv_loss_err'])
        return string

    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).

        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0 
        """
        if self.results != {}:
            output = [{'metric': 'avg_nndr', 'dim': 'p', 
                     'val': self.results['avg'], 
                     'err': self.results['err'], 
                     'n_val': self.results['avg'], 
                     'n_err': self.results['err'], 
                     }]
            if self.hout_data is not None:
                output.extend([{'metric': 'priv_loss_nndr', 'dim': 'p', 
                     'val': self.results['priv_loss'], 
                     'err': self.results['priv_loss_err'], 
                     'n_val': 1-abs(self.results['priv_loss']), 
                     'n_err': self.results['err'], 
                     }])
            return output
        else: pass


# Description: Kolmogorov–Smirnov test implementation 
# Author: Anton D. Lautrup
# Date: 21-08-2023

import numpy as np

from ..core.metric import MetricClass

from collections import Counter
from scipy.stats import permutation_test,ks_2samp

from ...utils.plot_metrics import plot_significantly_dissimilar_variables

def _total_variation_distance(x,y):
    """Function for calculating the TVD (KS statistic equivalent)"""
    X, Y = Counter(x), Counter(y)
    merged = X + Y

    return np.round(0.5*sum([abs(X[key]/len(x)-Y[key]/len(y)) for key in merged.keys()]),4)

# def _discrete_ks_statistic(x, y):
#     """Function for calculating the KS statistic"""
#     KstestResult = ks_2samp(x,y)
#     return np.round(KstestResult.statistic,4)

def _discrete_ks(x, y, n_perms=1000):
    """Function for doing permutation test of discrete values in the KS test"""
    res = permutation_test((x, y), _total_variation_distance, n_resamples=n_perms, vectorized=False, permutation_type='independent', alternative='greater')

    return res.statistic, res.pvalue

class KolmogorovSmirnovTest(MetricClass):
    """The Metric Class is an abstract class that interfaces with 
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    
    """

    def name() -> str:
        """ Name/keyword to reference the metric"""
        return 'ks_test'

    def type() -> str:
        """ Set to 'privacy' or 'utility' """
        return 'utility'

    def evaluate(self, sig_lvl=0.05, n_perms = 1000) -> float | dict:
        """Function for executing the Kolmogorov-Smirnov test.
    
        Returns:
            Avg. KS dist: dict  - holds avg. and standard error of the mean (SE)
            Avg. KS pval: dict  - holds avg. and SE
            num of sigs : int   - the number of significant tests at sig_lvl
            frac of sigs: float - the fraction of significant tests at sig_lvl   
        """
        n_dists, c_dists = [], []
        pvals = []
        sig_cols = []
        
        self.sig_lvl = sig_lvl

        for category in self.real_data.columns:
            R = self.real_data[category]
            F = self.synt_data[category]

            if category in self.cat_cols:
                statistic, pvalue = _discrete_ks(F,R,n_perms)
                c_dists.append(statistic)
                pvals.append(pvalue)
            else:
                KstestResult = ks_2samp(R,F)
                statistic, pvalue = KstestResult.statistic, KstestResult.pvalue
                n_dists.append(statistic)
                pvals.append(pvalue)
            if pvalue < sig_lvl:
                sig_cols.append(category)

        if n_dists == []: avg_ks = np.nan; err_ks = np.nan
        else: avg_ks = np.mean(n_dists); err_ks = np.std(n_dists,ddof=1)/np.sqrt(len(n_dists))

        if c_dists == []: avg_tvd = np.nan; err_tvd = np.nan
        else: avg_tvd = np.mean(c_dists); err_tvd = np.std(c_dists,ddof=1)/np.sqrt(len(c_dists))

        ### Calculate number of significant tests, and fraction of significant tests
        self.results = {'avg stat' : np.mean(n_dists+c_dists), 'stat err' : np.std(n_dists+c_dists,ddof=1)/np.sqrt(len(n_dists+c_dists)),
                        'avg ks'   : avg_ks, 'ks err'   : err_ks,
                        'avg tvd'  : avg_tvd, 'tvd err'  : err_tvd,
                        'avg pval' : np.mean(pvals), 'pval err' : np.std(pvals,ddof=1)/np.sqrt(len(pvals)),
                        'num sigs' : len(sig_cols),
                        'frac sigs': len(sig_cols)/len(pvals),
                        'sigs cols': sig_cols
                        }

        if (self.verbose and sig_cols != []): plot_significantly_dissimilar_variables(self.real_data, self.synt_data, sig_cols, self.cat_cols)
        return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval. 
|                                          :                    |"""
        R = self.results
        if self.results != {}:
            string = """\
| Kolmogorov–Smirnov / Total Variation Distance test            |
|   -> average combined statistic          :   %.4f  %.4f   |
|       -> avg. Kolmogorov–Smirnov dist.   :   %.4f  %.4f   |
|       -> avg. Total Variation Distance   :   %.4f  %.4f   |
|   -> average combined p-value            :   %.4f  %.4f   |
|       -> # significant tests at a=%.2f   :   %2d               |
|       -> fraction of significant tests   :   %.4f           |""" % (R['avg stat'], R['stat err'],
                                                                      R['avg ks'], R['ks err'],
                                                                      R['avg tvd'], R['tvd err'],
                                                                      R['avg pval'], R['pval err'], 
                                                                      self.sig_lvl, R['num sigs'],
                                                                      R['frac sigs'])
            return string
        else: pass

    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).

        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0
        """
        if self.results != {}:
            R = self.results

            return [{'metric': 'ks_tvd_stat', 'dim': 'u', 
                     'val': R['avg stat'], 
                     'err': R['stat err'], 
                     'n_val': 1-R['avg stat'], 
                     'n_err': R['stat err'], 
                     },
                     {'metric': 'frac_ks_sigs', 'dim': 'u', 
                     'val': R['frac sigs'], 
                     'n_val': 1-R['frac sigs'], 
                     }]
        else: pass



# Description: Membership Inference Attack based on classification
# Author: Tobias Hyrup
# Date: 2023-10-30

import numpy as np
import pandas as pd
from ..core.metric import MetricClass
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.metrics import precision_score, recall_score, f1_score
from sklearn.preprocessing import LabelEncoder


class MIAClassifier(MetricClass):
    """The Metric Class is an abstract class that interfaces with
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    self.analysis_target: variable name

    """

    def name() -> str:
        """Name/keyword to reference the metric"""
        return "att_discl"

    def type() -> str:
        """Set to 'privacy' or 'utility'"""
        return "privacy"

    def _minmaxscale(self, df: pd.DataFrame, include: list) -> pd.DataFrame:
        """Function for min-max scaling of a dataframe"""
        for column in df.columns:
            if column in include:
                df[column] = (df[column] - df[column].min()) / (
                    df[column].max() - df[column].min()
                )
        return df

    def _predict_cat_target(
        self,
        real: pd.DataFrame,
        syn: pd.DataFrame,
        target: str,
        hout: pd.DataFrame = None,
    ):
        syn_predictors = syn.loc[:, syn.columns != target]
        real_predictors = real.loc[:, real.columns != target]

        syn_target = syn[target]
        real_target = real[target]

        if hout is not None:
            hout_predictors = hout.loc[:, hout.columns != target]
            hout_target = hout[target]

        total_targets = (
            pd.concat([real_target, hout_target]) if hout is not None else real_target
        )
        total_predictors = (
            pd.concat([real_predictors, hout_predictors])
            if hout is not None
            else real_predictors
        )

        clf = RandomForestClassifier(n_estimators=100).fit(syn_predictors, syn_target)

        total_targets = (
            pd.concat([real_target, hout_target]) if hout is not None else real_target
        )
        total_predictors = (
            pd.concat([real_predictors, hout_predictors])
            if hout is not None
            else real_predictors
        )

        preds = clf.predict(total_predictors)
        accuracy = sum(preds == total_targets) / len(total_targets)
        precision = precision_score(
            total_targets, preds, average="macro", zero_division=0
        )
        recall = recall_score(total_targets, preds, average="macro", zero_division=0)
        f1 = f1_score(total_targets, preds, average="macro", zero_division=0)

        return accuracy, precision, recall, f1

    def _predict_num_target(
        self,
        real: pd.DataFrame,
        syn: pd.DataFrame,
        target: str,
        hout: pd.DataFrame = None,
        threshold: float = 1 / 30,
    ):
        syn_target = syn[target]
        real_target = real[target]
        syn_predictors = syn.loc[:, syn.columns != target]
        real_predictors = real.loc[:, real.columns != target]

        if hout is not None:
            hout_predictors = hout.loc[:, hout.columns != target]
            hout_target = hout[target]

        total_targets = (
            pd.concat([real_target, hout_target]) if hout is not None else real_target
        )
        total_predictors = (
            pd.concat([real_predictors, hout_predictors])
            if hout is not None
            else real_predictors
        )

        rf_regressor = RandomForestRegressor(n_estimators=100).fit(
            syn_predictors, syn_target
        )

        preds = rf_regressor.predict(total_predictors)

        # check if differens from preds to target_real is less than the given threshold
        preds = np.where(np.abs(preds - total_targets) <= threshold, 1, 0)
        total_targets = [1] * len(total_targets)

        accuracy = sum(preds == total_targets) / len(total_targets)
        precision = precision_score(
            total_targets, preds, average="macro", zero_division=0
        )
        recall = recall_score(total_targets, preds, average="macro", zero_division=0)
        f1 = f1_score(total_targets, preds, average="macro", zero_division=0)

        return accuracy, precision, recall, f1

    def evaluate(self, numerical_dist_thresh: float = 1 / 30, sensitive: list = None) -> float | dict:
        pre_results = {
            "accuracy": [],
            "precision": [],
            "recall": [],
            "f1": [],
        }

        # Scale the numeric attributes
        if self.hout_data is not None:
            combined_data = pd.concat(
                [self.real_data, self.synt_data, self.hout_data], ignore_index=True
            )
        else:
            combined_data = pd.concat(
                [self.real_data, self.synt_data], ignore_index=True
            )
        scaled = self._minmaxscale(combined_data.copy(deep=True), self.num_cols)

        real_scaled = scaled.iloc[: len(self.real_data)]
        syn_scaled = scaled.iloc[
            len(self.real_data) : len(self.real_data) + len(self.synt_data)
        ]
        hout_scaled = (
            scaled.iloc[len(self.real_data) + len(self.synt_data) :]
            if self.hout_data is not None
            else None
        )

        # Compute attribute disclosure for each attribute with maximum adversarial knowledge
        sensitive_columns = sensitive if sensitive is not None else self.real_data.columns
        for column in sensitive_columns:
            if column in self.cat_cols:
                accuracy, precision, recall, f1 = self._predict_cat_target(
                    real=real_scaled, syn=syn_scaled, hout=hout_scaled, target=column
                )

            else:
                accuracy, precision, recall, f1 = self._predict_num_target(
                    real=real_scaled,
                    syn=syn_scaled,
                    hout=hout_scaled,
                    target=column,
                    threshold=numerical_dist_thresh,
                )

            pre_results["accuracy"].append(accuracy)
            pre_results["precision"].append(precision)
            pre_results["recall"].append(recall)
            pre_results["f1"].append(f1)

        # Compute mean accuracy, precision, recall, and F1-score with accompanying standard errors
        accuracy = np.mean(pre_results["accuracy"])
        accuracy_se = np.std(pre_results["accuracy"], ddof=1) / np.sqrt(
            len(pre_results["accuracy"])
        )
        precision = np.mean(pre_results["precision"])
        precision_se = np.std(pre_results["precision"], ddof=1) / np.sqrt(
            len(pre_results["precision"])
        )

        recall = np.mean(pre_results["recall"])
        recall_se = np.std(pre_results["recall"], ddof=1) / np.sqrt(
            len(pre_results["recall"])
        )

        f1 = np.mean(pre_results["f1"])
        f1_se = np.std(pre_results["f1"], ddof=1) / np.sqrt(len(pre_results["f1"]))

        self.results = {
            "Attr Dis accuracy": accuracy,
            "Attr Dis accuracy se": accuracy_se,
            "Attr Dis precision": precision,
            "Attr Dis precision se": precision_se,
            "Attr Dis recall": recall,
            "Attr Dis recall se": recall_se,
            "Attr Dis macro F1": f1,
            "Attr Dis macro F1 se": f1_se,
        }

        return self.results

    def format_output(self) -> str:
        """Return string for formatting the output, when the
                metric is part of SynthEval.
        |                                          :                    |"""
        if self.hout_data is not None:
            string = """\
| Attr. disclosure risk (acc. with holdout):   %.4f  %.4f   |""" % (
            self.results["Attr Dis accuracy"],
            self.results["Attr Dis accuracy se"],
        )
        else:
            string = """\
| Attr. disclosure risk (accuracy)         :   %.4f  %.4f   |""" % (
            self.results["Attr Dis accuracy"],
            self.results["Attr Dis accuracy se"],
        )
        return string

    def normalize_output(self) -> list:
        """This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).

        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0
        """
        if self.results != {}:
            return [
                {
                    "metric": "att_discl_risk",
                    "dim": "p",
                    "val": self.results["Attr Dis accuracy"],
                    "err": self.results["Attr Dis accuracy se"],
                    "n_val": 1 - self.results["Attr Dis accuracy"],
                    "n_err": self.results["Attr Dis accuracy se"],
                }
            ]
        else:
            pass



# Description: Script with the preprocessing steps for metrics to work
# Author: Anton D. Lautrup
# Date: 16-11-2022

import numpy as np
import pandas as pd

from sklearn.preprocessing import OrdinalEncoder, MinMaxScaler


def stack(real, fake):
    """Function for stacking the real and fake dataframes and adding a column for keeping
    track of which is which. This is essentially to ease the use of seaborn plots hue.
    """
    real = pd.concat(
        (real.reset_index(), pd.DataFrame(np.ones(len(real)), columns=["real"])), axis=1
    )
    fake = pd.concat(
        (fake.reset_index(), pd.DataFrame(np.zeros(len(fake)), columns=["real"])),
        axis=1,
    )
    return pd.concat((real, fake), ignore_index=True)


class consistent_label_encoding:
    def __init__(
        self, real, fake, categorical_columns, nummerical_columns, hout=None
    ) -> None:
        assert (
            len(categorical_columns) > 0 or len(nummerical_columns) > 0
        ), "Either categorical or nummerical columns must be provided."

        joint_dataframe = pd.concat((real.reset_index(), fake.reset_index()), axis=0)
        if hout is not None:
            joint_dataframe = pd.concat(
                (joint_dataframe.reset_index(), hout.reset_index()), axis=0
            )

        if len(categorical_columns) > 0:
            self.encoder = OrdinalEncoder().fit(joint_dataframe[categorical_columns])
            self.cat_cols = categorical_columns
        else:
            self.cat_cols = None

        if len(nummerical_columns) > 0:
            self.num_encoder = MinMaxScaler().fit(joint_dataframe[nummerical_columns])
            self.num_cols = nummerical_columns
        else:
            self.num_cols = None

        # self.num_attribute_range = np.max(np.stack((np.array(np.ptp(joint_dataframe[nummerical_columns]),dtype=np.float64),np.ones(len(nummerical_columns)))),axis=0)
        pass

    def encode(self, data):
        data = data.copy()
        if self.cat_cols is not None:
            data[self.cat_cols] = self.encoder.transform(data[self.cat_cols]).astype(
                "int"
            )
        if self.num_cols is not None:
            data[self.num_cols] = self.num_encoder.transform(data[self.num_cols])
        return data

    def decode(self, data):
        data = data.copy()
        if self.cat_cols is not None:
            data[self.cat_cols] = self.encoder.inverse_transform(data[self.cat_cols])
        if self.num_cols is not None:
            data[self.num_cols] = self.num_encoder.inverse_transform(data[self.num_cols])
        return data


# Description: Template script for making new metric classes
# Author: Anton D. Lautrup
# Date: 21-08-2023

from .core.metric import MetricClass

class MetricClassName(MetricClass):
    """The Metric Class is an abstract class that interfaces with 
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    self.analysis_target: variable name

    self.verbose   : bool (mainly for supressing prints and plots)

    """

    def name() -> str:
        """ Name/keyword to reference the metric"""
        return 'temp'

    def type() -> str:
        """ Set to 'privacy' or 'utility' """
        return 'utility'

    def evaluate(self) -> float | dict:
        """ Function for evaluating the metric and returning raw results"""

        self.results = {}
        return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval. 
|                                          :                    |"""
        string = """\
|                                          :   %.4f  %.4f   |""" % (0.0000, 0.0000)
        return string

    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).
        This format is required for the metric to be used in the benchmark module. 
        
        The required format is:

        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0

        Error fields can be empty.

        dim, is for designating if metric covers utility ('u') or privacy ('p').

        [result] First set of val and err is the regular result found in the 
        results dictionary.

        [Benchmark] Second set is normalised to the zero-one interval so zero 
        represents the worst possible performance and one is the best possible 
        performance."""
        
        if self.results != {}:

            return [{'metric': 'placeholder', 'dim': 'u', 'val': 0.0, 'err': 0.0, 'n_val': 0.0, 'n_err': 0.0}]#, 'idx_val': 0.0, 'idx_err': 0.0}]
        else: pass


    ### Hooks (Extra functions, not required)
    def extra_formatted_output(self) -> str:
        """ Some metrics may output both privacy and utility results. For keeping 
        these results seperate in the console print, string output can be placed here 
        to be put in the end of the opposite console text output than the metric type 
        specified above.
|                                          :                    |""" 
        pass








# Description: Script for hosting the main framework
# Author: Anton D. Lautrup
# Date: 16-08-2023

import json
import glob
import time
import os

import numpy as np
import pandas as pd
from tqdm import tqdm
from typing import Literal
from pandas import DataFrame

from .metrics import load_metrics
from .utils.console_output import print_results_to_console
from .utils.preprocessing import consistent_label_encoding
from .utils.postprocessing import extremes_ranking, linear_ranking, quantile_ranking, summation_ranking
from .utils.variable_detection import get_cat_variables

loaded_metrics = load_metrics()
#print(loaded_metrics)

def _has_not_slash_backslash_or_dot(input_string):
    return not ('/' in input_string or '\\' in input_string or '.' in input_string)

class SynthEval():
    def __init__(self, 
                 real_dataframe: DataFrame, 
                 holdout_dataframe: DataFrame = None,
                 cat_cols: list = None,
                 nn_distance: Literal['gower', 'euclid', 'EXPERIMENTAL_gower'] = 'gower', 
                 unique_threshold: int = 10,
                 verbose: bool = True,
        ) -> None:
        """Primary object for accessing the SynthEval evaluation framework. Create with the real data used for training 
        and use either evaluate of benchmark methods for evaluating synthetic datasets.
        
        Parameters:
            real_dataframe      : real dataset, in dataframe format. 
            holdout_dataframe   : (optional) real data that was not used for training the generative model
            cat_cols            : (optional) complete list of categorical columns column names. 
            nn_distance         : {default= 'gower', 'euclid', 'EXPERIMENTAL_gower'} distance metric for NN distances.
            unique_threshold    : threshold of unique levels in non-object columns to be considered categoricals.    
            verbose             : flag fo printing to console and making figures
        """

        self.real = real_dataframe
        self.verbose = verbose

        if holdout_dataframe is not None:
            # Make sure columns and their order are the same.
            if len(real_dataframe.columns) == len(holdout_dataframe.columns):
                holdout_dataframe = holdout_dataframe[real_dataframe.columns.tolist()]
            assert real_dataframe.columns.tolist() == holdout_dataframe.columns.tolist(), 'Columns in real and houldout dataframe are not the same'

            self.hold_out = holdout_dataframe
        else:
            self.hold_out = None

        if cat_cols is None:
            cat_cols = get_cat_variables(real_dataframe, unique_threshold)
            if self.verbose: print('SynthEval: inferred categorical columns...')
        
        self.categorical_columns = cat_cols
        self.numerical_columns = [column for column in real_dataframe.columns if column not in cat_cols]

        self.nn_dist = nn_distance
        
        pass

    def _update_syn_data(self, synt):
        """Function for adding/updating the synthetic data"""
        self.synt = synt

        if len(self.real.columns) == len(synt.columns):
            synt = synt[self.real.columns.tolist()]
        assert self.real.columns.tolist() == synt.columns.tolist(), 'Columns in real and fake dataframe are not the same'
        if self.verbose: print('SynthEval: synthetic data read successfully')
        pass

    def display_loaded_metrics(self):
        """Utility function for getting an overview of the loaded modules and their keys."""
        print(loaded_metrics)
        pass

    def evaluate(self, synthetic_dataframe, analysis_target_var=None, presets_file=None, **kwargs):
        """Method for generating the SynthEval evaluation report on a synthetic dataset. Includes the metrics specified in the 
        presets file or through the keyword arguments. Returns a dataframe with the primary results, and prints to console if 
        verbose. The raw output can be accessed as a charateristic of the SynthEval object after running this method 
        (e.i. self._raw_results).
        
        Parameters:
            synthetic_dataframe     : synthetic dataset, in dataframe format. 
            analysis_target_var     : string column name of categorical variable to check.
            presets_file            : {default=None, 'full_eval', 'fast_eval', 'privacy'} or json file path.
            **kwargs                : keyword arguments for metrics e.g. ks_test={}, eps_risk={}, ...

        Returns:
            key_results : dataframe with the primary result(s) from each of the included metrics.
        """
        self._update_syn_data(synthetic_dataframe)
        
        loaded_preset = {}
        if presets_file is not None:
            ext = presets_file.split(".")[-1].lower()
            if _has_not_slash_backslash_or_dot(presets_file):
                with open(os.path.dirname(__file__)+'/presets/'+presets_file+'.json','r') as fp:
                    loaded_preset = json.load(fp)
            elif (ext == "json"):
                with open(presets_file,'r') as fp:
                    loaded_preset = json.load(fp)
            else:
                raise Exception("Error: unrecognised preset keyword or file format!")
        
        evaluation_config = {**loaded_preset, **kwargs}

        CLE = consistent_label_encoding(self.real, self.synt, self.categorical_columns, self.numerical_columns, self.hold_out)
        real_data = CLE.encode(self.real)
        synt_data = CLE.encode(self.synt)
        if self.hold_out is not None: hout_data = CLE.encode(self.hold_out)
        else: hout_data = None

        utility_output_txt = ''
        privacy_output_txt = ''

        methods = evaluation_config.keys()

        raw_results = {}
        key_results = None
        #scores = {'utility':{'val':[],'err':[]}, 'privacy':{'val':[],'err':[]}}
        pbar = tqdm(methods, disable= not self.verbose)
        for method in pbar:
            pbar.set_description(f'Syntheval: {method}')
            if method not in loaded_metrics.keys():
                print(f"Unrecognised keyword: {method}")
                #raise Exception(f"Unrecognised keyword: {method}")
                continue
            
            M = loaded_metrics[method](real_data, synt_data, hout_data, self.categorical_columns, self.numerical_columns, self.nn_dist, analysis_target_var, do_preprocessing=CLE, verbose=self.verbose)
            raw_results[method] = M.evaluate(**evaluation_config[method])

            string       = M.format_output()
            extra_string = M.extra_formatted_output()
            if string is not None:
                if loaded_metrics[method].type() == 'utility':
                    utility_output_txt += string + '\n'
                    if extra_string is not None: privacy_output_txt += extra_string + '\n'
                else:
                    privacy_output_txt += string + '\n'
                    if extra_string is not None: utility_output_txt += extra_string + '\n'

            normalized_result = M.normalize_output()
            if normalized_result is not None: 
                if key_results is None:
                    key_results = pd.DataFrame(M.normalize_output(), columns=['metric', 'dim', 'val','err','n_val','n_err'])#,'idx_val','idx_err'])
                else:
                    tmp_df = pd.DataFrame(M.normalize_output(), columns=['metric', 'dim', 'val','err','n_val','n_err'])#,'idx_val','idx_err'])
                    key_results = pd.concat((key_results,tmp_df), axis = 0).reset_index(drop=True)

        # scores = {'utility':{'val': key_results[key_results['dim'] == 'u']['idx_val'].dropna().tolist(),
        #                      'err': key_results[key_results['dim'] == 'u']['idx_err'].dropna().tolist()},
        #           'privacy':{'val': key_results[key_results['dim'] == 'p']['idx_val'].dropna().tolist(),
        #                      'err': key_results[key_results['dim'] == 'p']['idx_err'].dropna().tolist()}
        #                      }

        if self.verbose: print_results_to_console(utility_output_txt,privacy_output_txt)#,scores)

        if (kwargs != {} and self.verbose):
            with open('SE_config.json', "w") as json_file:
                json.dump(evaluation_config, json_file)

        self._raw_results = raw_results
        return key_results

    def benchmark(self, dict_or_path_to_syn_file_folder, analysis_target_var=None, presets_file=None, rank_strategy='linear', **kwargs):
        """Method for running SynthEval multiple times across all synthetic data files in a
        specified directory. Making a results file, and calculating rank-derived utility 
        and privacy scores.
        
        Parameters:
            dict_or_path_to_syn_file_folder : dict of dataframes or string like '/example/ex_data_dir/' to folder with datasets
            analysis_target_var             : string column name of categorical variable to check
            rank_strategy                   : {default='linear', 'normal', 'quantile', 'summation'} see descriptions below

        Returns:
            comb_df : dataframe with the metrics and their rank derived scores
            rank_df : dataframe with the ranks used to make the scores
            """

        # Part to avoid printing in the following and resetting to user preference after
        verbose_flag = False
        if self.verbose == True:
            verbose_flag = True
            self.verbose = False

        # Part to make sure input is the correct format
        if isinstance(dict_or_path_to_syn_file_folder, str):
            csv_files = sorted(glob.glob(dict_or_path_to_syn_file_folder + '*.csv'))
            
            df_dict = {}
            for file in csv_files: df_dict[file.split(os.path.sep)[-1].replace('.csv', '')] = pd.read_csv(file)
        elif isinstance(dict_or_path_to_syn_file_folder, dict):
            df_dict = dict_or_path_to_syn_file_folder
        else:
            raise Exception("Error: input was not instance of dictionary or filepath!")

        # Evaluate the datasets in parallel
        from joblib import Parallel, delayed
        res_list = Parallel(n_jobs=-2)(delayed(self.evaluate)(dataframe, analysis_target_var, presets_file, **kwargs) for dataframe in df_dict.values())
        
        results = {}
        for res, key in zip(res_list,list(df_dict.keys())): results[key] = res

        # Part to postprocess the results, format and rank them in a csv file.
        tmp_df = results[list(results.keys())[0]]

        utility_mets = tmp_df[tmp_df['dim'] == 'u']['metric'].tolist()
        privacy_mets = tmp_df[tmp_df['dim'] == 'p']['metric'].tolist()

        vals_df = pd.DataFrame(columns=tmp_df['metric'])
        errs_df = pd.DataFrame(columns=tmp_df['metric'])
        rank_df = pd.DataFrame(columns=tmp_df['metric'])

        for key, df in results.items():
            df = df.set_index('metric').T

            vals_df.loc[len(vals_df)] = df.loc['val']
            errs_df.loc[len(vals_df)] = df.loc['err']
            rank_df.loc[len(vals_df)] = df.loc['n_val']

        vals_df['dataset'] = list(results.keys())
        errs_df['dataset'] = list(results.keys())
        rank_df['dataset'] = list(results.keys())

        vals_df = vals_df.set_index('dataset')
        errs_df = errs_df.set_index('dataset')
        rank_df = rank_df.set_index('dataset')

        if rank_strategy == 'normal': rank_df = extremes_ranking(rank_df,utility_mets,privacy_mets)
        if rank_strategy == 'linear': rank_df = linear_ranking(rank_df,utility_mets,privacy_mets)
        if rank_strategy == 'quantile': rank_df = quantile_ranking(rank_df,utility_mets,privacy_mets)
        if rank_strategy == 'summation': rank_df = summation_ranking(rank_df,utility_mets,privacy_mets)

        comb_df = pd.DataFrame()
        for column in vals_df.columns:
            comb_df[(column, 'value')] = vals_df[column]
            comb_df[(column, 'error')] = errs_df[column]
        comb_df.columns = pd.MultiIndex.from_tuples(comb_df.columns)

        comb_df['rank'] = rank_df['rank']
        comb_df['u_rank'] = rank_df['u_rank']
        comb_df['p_rank'] = rank_df['p_rank']

        name_tag = str(int(time.time()))
        temp_df = comb_df.copy()
        temp_df.columns = ['_'.join(col) for col in comb_df.columns.values]
        vals_df.to_csv('SE_benchmark_results' +'_' +name_tag+ '.csv')
        temp_df.to_csv('SE_benchmark_ranking' +'_' +name_tag+ '.csv')

        self.verbose = verbose_flag
        return comb_df, rank_df


# Description: propensity mean squared error
# Author: Anton D. Lautrup
# Date: 23-08-2023

import numpy as np
import pandas as pd

from ..core.metric import MetricClass

from sklearn.model_selection import StratifiedKFold
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import StandardScaler

from sklearn.metrics import f1_score

from ...utils.preprocessing import stack

class PropensityMeanSquaredError(MetricClass):
    """The Metric Class is an abstract class that interfaces with 
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    self.analysis_target: variable name
    """

    def name() -> str:
        """ Name/keyword to reference the metric"""
        return 'p_mse'

    def type() -> str:
        """ Set to 'privacy' or 'utility' """
        return 'utility'

    def evaluate(self, k_folds=5, max_iter=100, solver='liblinear') -> float | dict:
        """Train a a discriminator to distinguish between real and fake data."""

        discriminator = LogisticRegression(max_iter=max_iter, solver=solver, random_state=42)
        Df = stack(self.real_data,self.synt_data).drop(['index'], axis=1)

        if len(self.num_cols) > 0:
            Df[self.num_cols] = StandardScaler().fit_transform(Df[self.num_cols])
        Xs, ys = Df.drop(['real'], axis=1), Df['real']

        ### Run 5-fold cross-validation
        kf = StratifiedKFold(n_splits=k_folds, shuffle=True, random_state=42)
        res, acc = [], []
        for train_index, test_index in kf.split(Xs, ys):
            x_train = Xs.iloc[train_index]
            x_test = Xs.iloc[test_index]
            y_train = ys.iloc[train_index]
            y_test = ys.iloc[test_index]

            mod = discriminator.fit(x_train,y_train)
            pred = mod.predict_proba(x_test)

            num_synths = len(y_test)-np.count_nonzero(y_test)
            res.append(np.mean((pred[:,0]-num_synths/len(y_test))**2))
            acc.append(f1_score(y_test,mod.predict(x_test),average='macro'))

        self.results = {'avg pMSE': np.mean(res), 
                        'pMSE err': np.std(res,ddof=1)/np.sqrt(len(res)),
                        'avg acc': np.mean(acc), 
                        'acc err': np.std(acc,ddof=1)/np.sqrt(len(acc))
                        }
        return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval. 
|                                          :                    |"""
        string = """\
| Propensity mean squared error (pMSE)     :   %.4f  %.4f   |
|   -> average pMSE classifier accuracy    :   %.4f  %.4f   |""" % (
        self.results['avg pMSE'],
        self.results['pMSE err'],
        self.results['avg acc'],
        self.results['acc err']
        )
        return string
    
    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).

        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0
        """
        if self.results != {}:
            return [{'metric': 'avg_pMSE', 'dim': 'u', 
                     'val': self.results['avg pMSE'], 
                     'err': self.results['pMSE err'], 
                     'n_val': 1-4*self.results['avg pMSE'], 
                     'n_err': 4*self.results['pMSE err'], 
                     }]
        else: pass



# Description: Template script for making new metric classes
# Author: Anton D. Lautrup
# Date: 18-08-2023

from pandas import DataFrame
from abc import ABC, abstractmethod

from ...utils.variable_detection import get_cat_variables
from ...utils.preprocessing import consistent_label_encoding

class MetricClass(ABC):
    """
    The Metric Class defines an abstract method that contains a skeleton of
    some evaluation metric algorithm
    """ 

    def __init__(
            self,
            real_data: DataFrame,
            synt_data: DataFrame,
            hout_data: DataFrame = None,
            cat_cols: list = None,
            num_cols: list = None,
            nn_dist: str = None,
            analysis_target : str = None,
            do_preprocessing: bool = True,
            verbose: bool = True
    ) -> None:
        
        if isinstance(do_preprocessing, (int, bool)) and do_preprocessing == True:
            if cat_cols is None:
                cat_cols = get_cat_variables(real_data, threshold=10)
                num_cols = [column for column in real_data.columns if column not in cat_cols]
                print('SynthEval: inferred categorical columns...')
                
            CLE = consistent_label_encoding(real_data, synt_data, cat_cols, num_cols, hout_data)
            real_data = CLE.encode(real_data)
            synt_data = CLE.encode(synt_data)
            if hout_data is not None: hout_data = CLE.encode(hout_data)
            self.encoder = CLE
        elif not isinstance(do_preprocessing, (int, bool)):
            self.encoder = do_preprocessing

        self.real_data = real_data
        self.synt_data = synt_data
        self.hout_data = hout_data
        self.cat_cols = cat_cols
        self.num_cols = num_cols

        self.nn_dist = nn_dist
        self.analysis_target = analysis_target

        self.results = {}

        self.verbose = verbose

        pass

    @staticmethod
    @abstractmethod
    def name() -> str:
        """name/keyword to reference the metric"""
        pass

    @staticmethod
    @abstractmethod
    def type() -> str:
        """privacy or utility"""
        pass

    @abstractmethod
    def evaluate(self) -> float | dict:
        """ Function for evaluating the metric"""
        pass

    @abstractmethod
    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval. 
|                                          :                    |       
        """
        pass

    @abstractmethod
    def normalize_output(self) -> dict:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).
        This format is required for the metric to be used in the benchmark module. 
        
        The required format is:

        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0

        Error fields can be empty.

        dim, is for designating if metric covers utility ('u') or privacy ('p').

        [result] First set of val and err is the regular result found in the 
        results dictionary.

        [Benchmark] Second set is normalised to the zero-one interval so zero 
        represents the worst possible performance and one is the best possible 
        performance."""
        pass
    
    ### Hooks
    def extra_formatted_output(self) -> tuple:
        """ Some metrics may output both privacy and utility results. For keeping 
        these results seperate in the console print, string output can be placed here 
        to be put in the end of the opposite console text output than the metric type 
        specified above.

        """
        pass


# Description: Script for keeping track of already calculated NNs
# Author: Anton D. Lautrup
# Date: 21-08-2023

# import gower

import numpy as np

from typing import Literal
from sklearn.neighbors import NearestNeighbors

def _create_matrix_with_ones(indices, num_rows):
    matrix = np.zeros((len(indices),num_rows), dtype=int)
    for i, index in enumerate(indices):
        matrix[i,index] = 1
    return matrix

### Custom Gower matrix
from sklearn.preprocessing import OrdinalEncoder
from scipy.spatial.distance import cdist

def _gower_matrix_sklearn(data_x, data_y=None, cat_features: list = None, weights=None, num_attribute_ranges=None, nums_metric: Literal['L1', 'EXP_L2'] = 'L1'):
    """Modified version of the python gower distance metric implementation
    url: https://pypi.org/project/gower/"""

    X = data_x
    if data_y is None: Y = data_x 
    else: Y = data_y 

    if not isinstance(X, np.ndarray): X = np.asarray(X)
    if not isinstance(Y, np.ndarray): Y = np.asarray(Y)

    x_n_rows, x_n_cols = X.shape
    y_n_rows, y_n_cols = Y.shape 
    
    out_shape = np.zeros((x_n_rows, y_n_rows), dtype=np.float32)

    ### Bit to infer, cat_features if nothing is supplied 
    if cat_features is None:
        if not isinstance(X, np.ndarray): 
            is_number = np.vectorize(lambda x: not np.issubdtype(x, np.number))
            cat_features = is_number(X.dtypes)    
        else:
            cat_features = np.zeros(x_n_cols, dtype=bool)
            for col in range(x_n_cols):
                if not np.issubdtype(type(X[0, col]), np.number):
                    cat_features[col]=True
    else:          
        cat_features = np.array(cat_features)

    ### Separate out weights
    if weights is None:
        weights = np.ones(X.shape[1])
            
    weights_cat = weights[cat_features]
    weights_num = weights[np.logical_not(cat_features)]

    ### Subsetting
    Z = np.concatenate((X,Y))
    
    x_index = range(0,x_n_rows)
    y_index = range(x_n_rows,x_n_rows+y_n_rows)
    
    Z_num = Z[:,np.logical_not(cat_features)]
    Z_cat = Z[:,cat_features]

    ### Make the denominator for the nummerical normalisation 
    if num_attribute_ranges is None:
        num_attribute_ranges = np.max(np.stack((np.array(np.ptp(Z_num,axis=0),dtype=np.float64),np.ones(len(weights_num)))),axis=0)

    X_num = Z_num[x_index,]
    Y_num = Z_num[y_index,]

    ### Do the nummerical step
    if not np.array_equal(cat_features,np.ones(X.shape[1])):
        if nums_metric == 'L1':
                nums_sum = cdist(X_num.astype(float), Y_num.astype(float), 'minkowski', p=1, w=(weights_num/num_attribute_ranges))

        elif nums_metric == 'EXP_L2':
                nums_sum = cdist(X_num.astype(float), Y_num.astype(float), 'minkowski', p=2, w=(weights_num/num_attribute_ranges**2))#/np.sqrt(len(weights_num))

        else: raise NotImplementedError("The keyword literal is not a valid!")
    else: nums_sum = out_shape
    
    ### Do the categorical step
    if not np.array_equal(cat_features,np.zeros(X.shape[1])):
        Z_cat_enc = OrdinalEncoder().fit_transform(Z_cat)

        X_cat = Z_cat_enc[x_index,]
        Y_cat = Z_cat_enc[y_index,]

        cat_sum = cdist(X_cat.astype(int),Y_cat.astype(int), 'hamming', w=weights_cat)*len(weights_cat)
    else: cat_sum = out_shape
    
    return (nums_sum+cat_sum)/weights.sum()



def _knn_distance(a, b, cat_cols, num, metric: Literal['gower', 'euclid', 'EXPERIMENTAL_gower'] = 'gower', weights=None):
    def gower_knn(a, b, bool_cat_cols, gower_variant):
            """Function used for finding nearest neighbours"""
            d = []
            if np.array_equal(a,b):
                matrix = _gower_matrix_sklearn(a, cat_features=bool_cat_cols, weights=weights, nums_metric=gower_variant)+np.eye(len(a))
                for _ in range(num):
                    d.append(matrix.min(axis=1))
                    matrix += _create_matrix_with_ones(matrix.argmin(axis=1,keepdims=True),len(a))
            else:
                matrix = _gower_matrix_sklearn(a, b, cat_features=bool_cat_cols, weights=weights, nums_metric=gower_variant)
                for _ in range(num):
                    d.append(matrix.min(axis=1))
                    matrix += _create_matrix_with_ones(matrix.argmin(axis=1,keepdims=True),len(b))
            return d

    def eucledian_knn(a, b):
            """Function used for finding nearest neighbours"""
            d = []
            nn = NearestNeighbors(n_neighbors=num+1, metric_params={'w':weights}) #TODO: add num_att_range here as well
            if np.array_equal(a,b):
                nn.fit(a)
                dists, _ = nn.kneighbors(a)
                for i in range(num):
                    d.append(dists[:,1+i])
            else:
                nn.fit(b)
                dists, _ = nn.kneighbors(a)
                for i in range(num):
                    d.append(dists[:,i])
            return d

    if metric=='gower' or metric=='EXPERIMENTAL_gower':
        bool_cat_cols = [col1 in cat_cols for col1 in a.columns]
        num_cols = [col2 for col2 in a.columns if col2 not in cat_cols]
        a[num_cols] = a[num_cols].astype("float")
        b[num_cols] = b[num_cols].astype("float")
        if metric=='gower': return gower_knn(a,b,bool_cat_cols, gower_variant = 'L1')
        else: return gower_knn(a,b,bool_cat_cols, gower_variant='EXP_L2')
    if metric=='euclid':
        return eucledian_knn(a,b)
    else: raise Exception("Unknown metric; options are 'gower' or 'euclid'")



# Description: Mutual information metric and plot
# Author: Anton D. Lautrup
# Date: 21-08-2023

import numpy as np
import pandas as pd

from ..core.metric import MetricClass

from ...utils.plot_metrics import plot_matrix_heatmap
from sklearn.metrics import normalized_mutual_info_score

def _pairwise_attributes_mutual_information(data):
    """Compute normalized mutual information for all pairwise attributes.

    Elements borrowed from: 
    Ping H, Stoyanovich J, Howe B. DataSynthesizer: privacy-preserving synthetic datasets. 2017
    Presented at: Proceedingsof the 29th International Conference on Scientific and Statistical Database Management; 2017; Chicago.
    [doi:10.1145/3085504.3091117]"""

    labs = sorted(data.columns)
    res = (normalized_mutual_info_score(data[cat1].astype(str),data[cat2].astype(str),average_method='arithmetic') for cat1 in labs for cat2 in labs)
    return pd.DataFrame(np.fromiter(res, dtype=float).reshape(len(labs),len(labs)), columns = labs, index = labs)

class MutualInformation(MetricClass):

    def name() -> str:
        """name/keyword to reference the metric"""
        return 'mi_diff'

    def type() -> str:
        """privacy or utility"""
        return 'utility'

    def evaluate(self, axs_lim=(0,1), axs_scale='Blues') -> float | dict:
        """ Function for evaluating the metric"""
        r_mi = _pairwise_attributes_mutual_information(self.real_data)
        f_mi = _pairwise_attributes_mutual_information(self.synt_data)

        mi_mat = r_mi - f_mi
        if self.verbose: plot_matrix_heatmap(mi_mat,'Mutual information matrix difference', 'mi', axs_lim, axs_scale)
        
        self.results = {'mutual_inf_diff': np.linalg.norm(mi_mat, ord='fro'),'mi_mat_dims': len(mi_mat)}
        return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval.        
        """
        string = """\
| Pairwise mutual information difference   :   %.4f           |""" % (self.results['mutual_inf_diff'])
        return string

    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).

        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0
        """
        if self.results != {}:
            n_elements = int(self.results['mi_mat_dims']*(self.results['mi_mat_dims']-1)/2)
            return [{'metric': 'mutual_inf_diff', 'dim': 'u', 
                     'val': self.results['mutual_inf_diff'], 
                     'n_val': 1-self.results['mutual_inf_diff']/n_elements, 
                    #  'idx_val': 1-np.tanh(self.results['mutual_inf_diff'])
                     }]
        else: pass


from sys import argv

from click import Path, argument, command, option

from . import SynthEval


@command()
@argument(
    "evaluate",
    required=0
)
@option(
    "--real-data-file",
    "-r",
    type=Path(exists=True),
    help="""Path to csv file with real data.""",
)
@option(
    "--synt-data-file",
    "-s",
    type=Path(exists=True),
    help="""Path to csv file with synthetic data.""",
)
@option(
    "--test-data-file",
    "-h",
    default=None,
    required=0,
    type=Path(exists=True),
    help="""Path to csv file with real data that wasn't used for training.""",
)
@option(
    "--evaluation-config",
    "-j",
    required=0,
    default='full_eval',
    type=Path(exists=True),
    help="""Name of preset file or filepath to custom json config file.""",
)
@option(
    "--category-labels",
    "-l",
    required=0,
    default=None,
    type=Path(exists=True),
    help="""Path to txt file with comma separated labels.""",
)
@option(
    "--class-label",
    "-c",
    default=None,
    type=str,
    help="""Label to use for prediction usability and coloring on plots."""
)
def cli(evaluate,real_data_file,synt_data_file,test_data_file,evaluation_config,category_labels,class_label):
    from syntheval import SynthEval
    ext = real_data_file.split(".")[-1].lower()
    if ext == "csv":
        import pandas as pd
        df_real = pd.read_csv(real_data_file)
        df_fake = pd.read_csv(synt_data_file)
        if test_data_file is not None:
            df_test = pd.read_csv(test_data_file)
        else: df_test = None

    if category_labels is not None:
        if category_labels.split(".")[-1].lower() == "txt":
            with open(category_labels, 'r') as file:
                contents = file.read()
                category_labels = contents.split(',')
        else:
            raise RuntimeError("Please provide category labels in a comma separated txt file")

    evaluator = SynthEval(df_real, holdout_dataframe=df_test, cat_cols=category_labels)
    evaluator.evaluate(df_fake, class_label, evaluation_config)
    
#cli()


# Description: Metric implementation of the classification accuracy difference.
# Author: Anton D. Lautrup
# Date: 22-08-2023

import numpy as np
from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import f1_score
from sklearn.model_selection import KFold
from sklearn.tree import DecisionTreeClassifier
from tqdm import tqdm

from ..core.metric import MetricClass


def class_test(real_models, fake_models, real, fake, test, F1_type):
    """Function for running a training session and getting predictions 
    on the SciPy model provided, and data."""
    res = []
    for r_mod, f_mod in zip(real_models, fake_models):
        r_mod.fit(real[0],real[1])
        f_mod.fit(fake[0],fake[1])

        pred_real = r_mod.predict(test[0])
        pred_fake = f_mod.predict(test[0])

        f1_real = f1_score(test[1],pred_real,average=F1_type)
        f1_fake = f1_score(test[1],pred_fake,average=F1_type)

        res.append([f1_real, f1_fake])
    return np.array(res).T

class ClassificationAccuracy(MetricClass):
    """The Metric Class is an abstract class that interfaces with 
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    self.analysis_target: variable name
    
    """

    def name() -> str:
        """ Name/keyword to reference the metric"""
        return 'cls_acc'

    def type() -> str:
        """ Set to 'privacy' or 'utility' """
        return 'utility'

    def evaluate(self, F1_type='micro',k_folds=5) -> dict:
        """ Function for training classifiers"""
        if (self.analysis_target is not None and self.analysis_target in self.cat_cols):
            real_x, real_y = self.real_data.drop([self.analysis_target], axis=1), self.real_data[self.analysis_target]
            fake_x, fake_y = self.synt_data.drop([self.analysis_target], axis=1), self.synt_data[self.analysis_target]

            ### Build Classification the models            
            R_DT, F_DT = DecisionTreeClassifier(max_depth=15,random_state=42), DecisionTreeClassifier(max_depth=15,random_state=42)
            R_AB, F_AB = AdaBoostClassifier(n_estimators=10,algorithm='SAMME',learning_rate=1,random_state=42),AdaBoostClassifier(n_estimators=10,algorithm='SAMME',learning_rate=1,random_state=42)
            R_RF, F_RF = RandomForestClassifier(n_estimators=10,max_depth=15, random_state=42),RandomForestClassifier(n_estimators=10,max_depth=15, random_state=42)
            R_LG, F_LG = LogisticRegression(multi_class='auto', solver='saga', max_iter=5000, random_state=42), LogisticRegression(multi_class='auto', solver='saga', max_iter=5000, random_state=42)

            real_models = [R_DT, R_AB, R_RF, R_LG]
            fake_models = [F_DT, F_AB, F_RF, F_LG]

            ### Run 5-fold cross-validation
            kf = KFold(n_splits=k_folds)
            res = []
            smol = real_y if len(real_y)<len(fake_y) else fake_y
            for train_index, test_index in tqdm(kf.split(smol), desc='cls_acc', total=k_folds, disable = not self.verbose):
                real_x_train = real_x.iloc[train_index]
                real_x_test = real_x.iloc[test_index]
                real_y_train = real_y.iloc[train_index]
                real_y_test = real_y.iloc[test_index]
                fake_x_train = fake_x.iloc[train_index]
                fake_y_train = fake_y.iloc[train_index]

                res.append(class_test(real_models,fake_models,[real_x_train, real_y_train],
                                                            [fake_x_train, fake_y_train],
                                                            [real_x_test, real_y_test],F1_type))
            
            self.results = {}

            class_avg = np.mean(res,axis=0)
            class_err = np.std(res,axis=0,ddof=1)/np.sqrt(k_folds)
            class_diff = np.abs(class_avg[0,:]-class_avg[1,:])
            class_diff_err = np.sqrt(class_err[0,:]**2+class_err[1,:]**2)

            self.k_folds = k_folds
            self.class_avg = class_avg

            self.results['avg diff'] = np.mean(class_diff)
            self.results['avg diff err'] = 0.25*np.sqrt(sum(class_diff_err**2))
            self.results['diffs'] = class_diff
            self.results['diffs err'] = class_diff_err

            if self.hout_data is not None:
                holdout_res = class_test(real_models, fake_models, [real_x, real_y],
                                                                    [fake_x, fake_y],
                                                                    [self.hout_data.drop([self.analysis_target],axis=1), 
                                                                    self.hout_data[self.analysis_target]],
                                                                    F1_type)
                
                holdout_diff = np.abs(holdout_res[0,:]-holdout_res[1,:])

                self.holdout_res = holdout_res

                self.results['avg diff hout'] = np.mean(holdout_diff)
                self.results['avg diff err hout'] = np.std(holdout_diff,ddof=1)/np.sqrt(4)
                self.results['diffs hout'] = holdout_diff

            return self.results

        elif self.analysis_target is None: 
            print('Error: Classification accuracy metric did not run, analysis class variable not set!')
            pass
        else:
            print('Error: Classification accuracy metric did not run, provided class not in list of categoricals!')
            pass
        

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval. 
|                                          :                    |"""
        if self.results == {}:
            string = """\
| Classification accuracy [FAILED TO RUN]  :   -.----  -.----   |"""
        else:
            avgs = self.class_avg
            diff = self.results['diffs']
            err = self.results['diffs err']
            string = """\
+---------------------------------------------------------------+

Classification accuracy test     
avg. of %d-fold cross val.:
classifier model             acc_r   acc_f    |diff|  error
+---------------------------------------------------------------+
| DecisionTreeClassifier  :   %.4f  %.4f   %.4f  %.4f   |
| AdaBoostClassifier      :   %.4f  %.4f   %.4f  %.4f   | 
| RandomForestClassifier  :   %.4f  %.4f   %.4f  %.4f   |
| LogisticRegression      :   %.4f  %.4f   %.4f  %.4f   |
+---------------------------------------------------------------+
| Average                 :   %.4f  %.4f   %.4f  %.4f   |\n""" % ( self.k_folds,
avgs[0,0],avgs[1,0],diff[0],err[0], 
avgs[0,1],avgs[1,1],diff[1],err[1],
avgs[0,2],avgs[1,2],diff[2],err[2],
avgs[0,3],avgs[1,3],diff[3],err[3],
np.mean(avgs[0,:]),np.mean(avgs[1,:]),self.results['avg diff'], self.results['avg diff err']
)
        if (self.results != {} and self.hout_data is not None):
            hres = self.holdout_res
            hdiff = self.results['diffs hout']
            string += """\
+---------------------------------------------------------------+

hold out data results:
+---------------------------------------------------------------+
| DecisionTreeClassifier  :   %.4f  %.4f   %.4f           |
| AdaBoostClassifier      :   %.4f  %.4f   %.4f           |
| RandomForestClassifier  :   %.4f  %.4f   %.4f           |
| LogisticRegression      :   %.4f  %.4f   %.4f           |
+---------------------------------------------------------------+
| Average                 :   %.4f  %.4f   %.4f  %.4f   |""" % (
hres[0,0], hres[1,0], hdiff[0],
hres[0,1], hres[1,1], hdiff[1],
hres[0,2], hres[1,2], hdiff[2],
hres[0,3], hres[1,3], hdiff[3],
np.mean(hres[:,0]), np.mean(hres[:,1]), self.results['avg diff hout'], self.results['avg diff err hout']
)
#         else:
#             string += """\
# +---------------------------------------------------------------+"""
        return string

    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).
        
        The required format is:
        metric  val  err  n_val  n_err idx_val idx_err
            name1  0.0  0.0    0.0    0.0    None    None
            name2  0.0  0.0    0.0    0.0    0.0     0.0
        """
        if self.results !={}:
            # val_non_lin     = np.exp(-8*self.results['avg diff'])
            # val_non_lin_err = 8*val_non_lin*self.results['avg diff err']

            output = [{'metric': 'cls_F1_diff', 'dim': 'u',
                       'val': self.results['avg diff'], 
                       'err': self.results['avg diff err'], 
                       'n_val': 1-self.results['avg diff'], 
                       'n_err': self.results['avg diff err'], 
                    #    'idx_val': val_non_lin, 
                    #    'idx_err': val_non_lin_err
                       }]
            if (self.hout_data is not None):
                # val_non_lin_2       = np.exp(-8*self.results['avg diff hout'])
                # val_non_lin_err_2   = 8*val_non_lin_2*self.results['avg diff err hout']

                output.extend([{'metric': 'cls_F1_diff_hout', 'dim': 'u',
                       'val': self.results['avg diff hout'], 
                       'err': self.results['avg diff err hout'], 
                       'n_val': 1-self.results['avg diff hout'], 
                       'n_err': self.results['avg diff err hout'], 
                    #    'idx_val': val_non_lin_2, 
                    #    'idx_err': val_non_lin_err_2
                       }])
            return output
        else: pass


# Description: Implementation of auroc metric and plot of roc curves
# Author: Anton D. Lautrup
# Date: 07-11-2023

import numpy as np
import pandas as pd

from ..core.metric import MetricClass
from ...utils.plot_metrics import plot_roc_curves

from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import roc_curve, auc
from sklearn.utils import resample

class MetricClassName(MetricClass):
    """The Metric Class is an abstract class that interfaces with 
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    self.analysis_target: variable name

    self.verbose   : bool (mainly for supressing prints and plots)

    """

    def name() -> str:
        """ Name/keyword to reference the metric"""
        return 'auroc_diff'

    def type() -> str:
        """ Set to 'privacy' or 'utility' """
        return 'utility'

    def evaluate(self, model = 'log_reg', num_subsamples = 1) -> float | dict:
        """ Metric that calculates the AUROC difference between a Random Forest model trained on 
        real data and one trained on fake data. Also plots the ROC curves if verbose
        
        model: 'log_reg' or 'rf_cls'
        
        """
        try:
            assert self.analysis_target is not None
            assert len(pd.unique(self.real_data[self.analysis_target])) == 2
            assert self.hout_data is not None
        except AssertionError:
            print(" Warning: AUROC metric did not run, analysis target variable did not have appropriate number levels (i.e. 2) or test data was not supplied!")
            pass
        else:
            real_x, real_y = self.real_data.drop([self.analysis_target], axis=1), self.real_data[self.analysis_target]
            fake_x, fake_y = self.synt_data.drop([self.analysis_target], axis=1), self.synt_data[self.analysis_target]

            hout_x, hout_y = self.hout_data.drop([self.analysis_target], axis=1), self.hout_data[self.analysis_target]
            if model == 'rf_cls':
                model1 = RandomForestClassifier(random_state=42)
                model2 = RandomForestClassifier(random_state=42)
            elif model == 'log_reg':
                model1 = LogisticRegression(random_state=42, max_iter=100)
                model2 = LogisticRegression(random_state=42, max_iter=100)
            else:
                print(f" Error: Unrecognised model name '{model}'!")
                pass

            roc_curves_real = []
            roc_curves_fake = []
            for i in range(num_subsamples):
                if num_subsamples != 1:
                    real_x_sub, real_y_sub = resample(real_x, real_y, n_samples=int(len(real_x)/2), random_state=i)
                    fake_x_sub, fake_y_sub = resample(fake_x, fake_y, n_samples=int(len(fake_x)/2))
                else:
                    real_x_sub, real_y_sub = real_x, real_y
                    fake_x_sub, fake_y_sub = fake_x, fake_y
                
                model1.fit(real_x_sub, real_y_sub)
                model2.fit(fake_x_sub, fake_y_sub)
                y1_probs = model1.predict_proba(hout_x)[:, 1]
                y2_probs = model2.predict_proba(hout_x)[:, 1]
                
                # Calculate ROC curve for the subsampled model
                fpr1, tpr1, _ = roc_curve(hout_y, y1_probs)
                fpr2, tpr2, _ = roc_curve(hout_y, y2_probs)

                roc_curves_real.append((fpr1, tpr1))
                roc_curves_fake.append((fpr2, tpr2))
            
            mean_fpr = np.linspace(0, 1, len(fpr1))

            tprs_real, tprs_fake = [], []

            for fpr, tpr in roc_curves_real:
                tprs_real.append(np.interp(mean_fpr, fpr, tpr))

            mean_tpr_real = np.mean(tprs_real, axis=0)
            std_tpr_real = np.std(tprs_real, axis=0)

            for fpr, tpr in roc_curves_fake:
                tprs_fake.append(np.interp(mean_fpr, fpr, tpr))

            mean_tpr_fake = np.mean(tprs_fake, axis=0)
            std_tpr_fake = np.std(tprs_fake, axis=0)

            # Calculate AUROC for the mean ROC curve
            roc_auc_mean_real = auc(mean_fpr, mean_tpr_real)
            roc_auc_mean_fake = auc(mean_fpr, mean_tpr_fake)

            if self.verbose: plot_roc_curves([mean_fpr, mean_tpr_real, roc_auc_mean_real], 
                                             [mean_fpr, mean_tpr_real, std_tpr_real], 
                                             [mean_fpr, mean_tpr_fake, roc_auc_mean_fake],
                                             [mean_fpr, mean_tpr_fake, std_tpr_fake],
                                             model, 'roc_curves')

            self.results = {'model': model, 'auroc_diff': abs(roc_auc_mean_real - roc_auc_mean_fake)}
            return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval. 
|                                          :                    |"""
        try:
            assert self.analysis_target is not None
            assert len(pd.unique(self.real_data[self.analysis_target])) == 2
            assert self.hout_data is not None
        except AssertionError:
            pass
        else:
            string = """\
| prediction AUROC difference (%7s)    :   %.4f           |""" % (self.results['model'], self.results['auroc_diff'])
            return string

    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).

        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0
        """
        if self.results != {}:
            # val_non_lin = np.exp(-10*abs(self.results['auroc_diff']))
            return [{'metric': 'auroc', 'dim': 'u', 
                     'val': self.results['auroc_diff'], 
                     'n_val': 1-self.results['auroc_diff'], 
                    #  'idx_val': val_non_lin, 
                     }]
        else: pass


# Description: Principal component analysis plot
# Author: Anton D. Lautrup
# Date: 23-08-2023

import pandas as pd
import numpy as np

from typing import Literal

from ..core.metric import MetricClass

from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

from ...utils.plot_metrics import plot_principal_components, plot_own_principal_component_pairplot
from ...utils.preprocessing import stack

class PrincipalComponentAnalysis(MetricClass):
    """The Metric Class is an abstract class that interfaces with 
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    self.analysis_target: variable name

    """

    def name() -> str:
        """ Name/keyword to reference the metric"""
        return 'pca'

    def type() -> str:
        """ Set to 'privacy' or 'utility' """
        return 'utility'

    def evaluate(self, num_components = 2, preprocess: Literal['mean', 'std'] = 'mean') -> float | dict:
        """ Function for evaluating the metric
        
        Args:
        num_components : int
            Number of principal components to plot, the metrics are based on the first two regardless
        preprocess : str
            Preprocessing method for the data, either 'mean' or 'std'
        """

        try:
            assert ((not self.verbose or (self.analysis_target is not None and self.analysis_target in self.cat_cols)) and len(self.num_cols)>=2)
        except AssertionError:
            if len(self.num_cols)<2:
                print('Error: Principal component analysis did not run, too few nummerical attributes!')
            elif self.analysis_target is None: 
                print('Error: Principal component analysis did not run, analysis class variable not set!')
            else:
                print('Error: Principal component analysis did not run, provided class not in list of categoricals!')
            pass
        else:
            if preprocess == 'mean':
                real_data = self.encoder.decode(self.real_data)
                synt_data = self.encoder.decode(self.synt_data)
                r_scaled = real_data[self.num_cols] - real_data[self.num_cols].mean()
                f_scaled = synt_data[self.num_cols] - synt_data[self.num_cols].mean()
            else:
                r_scaled = StandardScaler().fit_transform(self.real_data[self.num_cols])
                f_scaled = StandardScaler().fit_transform(self.synt_data[self.num_cols])

            pca = PCA(n_components=num_components)
            r_pca = pca.fit_transform(r_scaled)
            f_pca = pca.transform(f_scaled)

            labels = [ f"PC {i+1} ({var:.1f}%)" for i, var in enumerate(pca.explained_variance_ratio_ * 100)]

            synt_pca = PCA(n_components=num_components)
            s_pca = synt_pca.fit_transform(f_scaled)

            var_difference = sum(abs(pca.explained_variance_ratio_- synt_pca.explained_variance_ratio_))

            len_r = np.sqrt(pca.components_[0].dot(pca.components_[0]))
            len_f = np.sqrt(synt_pca.components_[0].dot(synt_pca.components_[0]))

            angle_diff = min([np.arccos(pca.components_[0] @ (s*synt_pca.components_[0])) for s in [1,-1]])/(len_r*len_f)

            self.results = {'exp_var_diff': var_difference/2, 'comp_angle_diff': 2*angle_diff/np.pi}
            if len(self.num_cols) > 5 :
                r_sort = np.argsort(abs(pca.components_[0]))[::-1]
                s_sort = np.argsort(abs(synt_pca.components_[0]))[::-1]
                self.results['pca_max_cont_real'] = [self.num_cols[i] for i in r_sort[:5]]
                self.results['pca_max_cont_synt'] = [self.num_cols[i] for i in s_sort[:5]]
            
            r_pca = pd.DataFrame(r_pca,columns=labels)
            f_pca = pd.DataFrame(f_pca,columns=labels)
            s_pca = pd.DataFrame(s_pca,columns=labels)

            if self.verbose: 
                r_pca['target'] = self.real_data[self.analysis_target]
                f_pca['target'] = self.synt_data[self.analysis_target]
                plot_principal_components(r_pca,f_pca)
                plot_own_principal_component_pairplot(stack(r_pca,s_pca))
            return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval. 
|                                          :                    |"""
        if self.results != {}:
            string = """\
| PCA difference in eigenvalues (exp. var.):   %.4f           |
| PCA angle diff. between eigenvectors     :   %.4f           |""" % (self.results['exp_var_diff'], 
                                                                      self.results['comp_angle_diff'])
            return string
        else: pass

    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).

        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0
        """
        if self.results != {}:
            return [{'metric': 'pca_eigval_diff', 'dim': 'u', 
                     'val': self.results['exp_var_diff'], 
                     'n_val': 1-self.results['exp_var_diff'] 
                     },
                     {'metric': 'pca_eigvec_ang', 'dim': 'u', 
                     'val': self.results['comp_angle_diff'], 
                     'n_val': 1-self.results['comp_angle_diff'] 
                     }]
        else: pass









# Description: Dimensionwise means difference
# Author: Anton D. Lautrup
# Date: 23-08-2023

import numpy as np
import pandas as pd

from ..core.metric import MetricClass

from scipy.stats import sem

from ...utils.plot_metrics import plot_dimensionwise_means

class MetricClassName(MetricClass):
    """The Metric Class is an abstract class that interfaces with 
    SynthEval. When initialised the class has the following attributes:

    Attributes:
    self.real_data : DataFrame
    self.synt_data : DataFrame
    self.hout_data : DataFrame
    self.cat_cols  : list of strings
    self.num_cols  : list of strings

    self.nn_dist   : string keyword
    self.analysis_target: variable name
    """

    def name() -> str:
        """ Name/keyword to reference the metric"""
        return 'dwm'

    def type() -> str:
        """ Set to 'privacy' or 'utility' """
        return 'utility'

    def evaluate(self) -> float | dict:
        """Function for calculating DWM, plotting an appropriate diagram"""
        try:
            assert len(self.num_cols) > 0
        except AssertionError:
            print(" Warning: No nummerical attributes provided for dimensionwise means metric.")
            return {}
        else:
            real_data = self.real_data[self.num_cols]
            synt_data = self.synt_data[self.num_cols]

            dim_means = np.array([np.mean(real_data,axis=0),np.mean(synt_data,axis=0)]).T
            means_diff = dim_means[:,0]-dim_means[:,1]
            
            mean_errors = np.array([sem(real_data),sem(synt_data)]).T
            diff_error = np.sqrt(np.sum(mean_errors**2,axis=1))

            if self.verbose: plot_dimensionwise_means(dim_means, mean_errors, self.num_cols)
            self.results = {'avg': np.mean(abs(means_diff)), 'err': np.sqrt(sum(diff_error**2))/len(diff_error)}
            return self.results

    def format_output(self) -> str:
        """ Return string for formatting the output, when the
        metric is part of SynthEval. 
|                                          :                    |"""
        if self.results != {}:
            string = """\
| Average dimensionwise means diff. (nums) :   %.4f  %.4f   |""" % (self.results['avg'], self.results['err'])
            return string
        else: pass

    def normalize_output(self) -> list:
        """ This function is for making a dictionary of the most quintessential
        nummerical results of running this metric (to be turned into a dataframe).

        The required format is:
        metric  dim  val  err  n_val  n_err
            name1  u  0.0  0.0    0.0    0.0
            name2  p  0.0  0.0    0.0    0.0
        """
        if self.results != {}:
            return [{'metric': 'avg_dwm_diff', 'dim': 'u', 
                     'val': self.results['avg'], 
                     'err': self.results['err'], 
                     'n_val': 1-self.results['avg'], 
                     'n_err': self.results['err'], 
                     }]
        else: pass

